/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_misc.c

#include "quakedef.h"

cvar_t r_showtris = {"r_showtris", "0"};

// johnfitz -- new cvars
extern cvar_t r_clearcolor;
extern cvar_t r_drawflat;
extern cvar_t r_flatlightstyles;
extern cvar_t gl_fullbrights;
extern cvar_t gl_overbright;
extern cvar_t gl_overbright_models;
extern cvar_t r_waterwarp;
extern cvar_t r_oldskyleaf;
extern cvar_t r_drawworld;
extern cvar_t r_showbboxes;
extern cvar_t r_lerpmodels;
extern cvar_t r_lerpmove;
extern cvar_t r_nolerp_list;
// johnfitz

extern cvar_t r_lockpvs;
extern cvar_t r_lockfrustum;

extern gltexture_t *playertextures[MAX_SCOREBOARD]; // johnfitz

void R_NoLerpList_f (void); // johnfitz


cvar_t gl_src_blend = {"gl_src_blend", "0"};
cvar_t gl_dst_blend = {"gl_dst_blend", "0"};


void GL_BlendTest (void)
{
	GLenum sfactors[] = {GL_ZERO, GL_ONE, GL_DST_COLOR, GL_ONE_MINUS_DST_COLOR, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA, GL_SRC_ALPHA_SATURATE};
	GLenum dfactors[] = {GL_ZERO, GL_ONE, GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA};

	if (gl_src_blend.value < 0) Cvar_SetValue ("gl_src_blend", 0);
	if (gl_src_blend.value > 8) Cvar_SetValue ("gl_src_blend", 8);

	if (gl_dst_blend.value < 0) Cvar_SetValue ("gl_dst_blend", 0);
	if (gl_dst_blend.value > 7) Cvar_SetValue ("gl_dst_blend", 7);

	glBlendFunc (sfactors[(int) gl_src_blend.value], dfactors[(int) gl_dst_blend.value]);
}


gltexture_t *r_whitetexture = NULL;

#define ALPHATYPE_ALIAS		1
#define ALPHATYPE_BRUSH		2
#define ALPHATYPE_SPRITE	3
#define ALPHATYPE_LIQUID	4
#define ALPHATYPE_PARTICLE	5
#define ALPHATYPE_FENCE		6
#define ALPHATYPE_SURFACE	7
#define ALPHATYPE_CORONA	8

typedef struct alphalist_s
{
	union
	{
		entity_t *ent;
		msurface_t *surf;
		particle_type_t *particles;
		dlight_t *light;
	};

	xcommand_t setupfunc;
	xcommand_t takedownfunc;

	float *matrix;

	int type;
	float dist;
} alphalist_t;


#define MAX_ALPHALIST 65536

alphalist_t *r_alphalist[MAX_ALPHALIST] = {NULL};
int r_numalphalist = 0;

float R_GetDist (float *origin)
{
	// no need to sqrt these as all we're concerned about is relative distances
	// (if x < y then sqrt (x) is also < sqrt (y))
	return
	(
		(origin[0] - r_origin[0]) * (origin[0] - r_origin[0]) +
		(origin[1] - r_origin[1]) * (origin[1] - r_origin[1]) +
		(origin[2] - r_origin[2]) * (origin[2] - r_origin[2])
	);
}


void R_AlphaListBeginMap (void)
{
	int i;

	// if we ever move this from Hunk memory we'll need to free it too...
	for (i = 0; i < MAX_ALPHALIST; i++)
		r_alphalist[i] = NULL;

	r_numalphalist = 0;
}


void R_AlphaListBeginFrame (void)
{
	r_numalphalist = 0;
}


alphalist_t *R_GrabAlphaSlot (void)
{
	if (r_numalphalist >= MAX_ALPHALIST) return NULL;

	if (!r_alphalist[r_numalphalist])
		r_alphalist[r_numalphalist] = (alphalist_t *) Hunk_Alloc (sizeof (alphalist_t));

	if (!r_alphalist[r_numalphalist]) return NULL;

	r_alphalist[r_numalphalist]->setupfunc = NULL;
	r_alphalist[r_numalphalist]->takedownfunc = NULL;

	r_numalphalist++;
	return r_alphalist[r_numalphalist - 1];
}


void R_CoronasBegin (void);
void R_CoronasEnd (void);
void R_RenderCorona (dlight_t *light);

void R_AddCoronaToAlphaList (dlight_t *l)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_CoronasBegin;
	al->takedownfunc = R_CoronasEnd;
	al->type = ALPHATYPE_CORONA;
	al->dist = R_GetDist (l->origin);
	al->light = l;
}


void R_ParticlesBegin (void);
void R_ParticlesEnd (void);

void R_AddParticlesToAlphaList (particle_type_t *pt)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_ParticlesBegin;
	al->takedownfunc = R_ParticlesEnd;
	al->type = ALPHATYPE_PARTICLE;
	al->dist = R_GetDist (pt->spawnorg);
	al->particles = pt;
}


void R_LiquidBegin (void);
void R_LiquidEnd (void);
void R_SpriteBegin (void);
void R_SpriteEnd (void);

float *R_TransformMidPoint (float *midpoint, float *matrix)
{
	static float transformed[3] = {0, 0, 0};

	if (matrix)
	{
		transformed[0] = midpoint[0] * matrix[0] + midpoint[1] * matrix[4] + midpoint[2] * matrix[8] + matrix[12];
		transformed[1] = midpoint[0] * matrix[1] + midpoint[1] * matrix[5] + midpoint[2] * matrix[9] + matrix[13];
		transformed[2] = midpoint[0] * matrix[2] + midpoint[1] * matrix[6] + midpoint[2] * matrix[10] + matrix[14];
	}
	else
	{
		transformed[0] = midpoint[0];
		transformed[1] = midpoint[1];
		transformed[2] = midpoint[2];
	}

	return transformed;
}


void R_AddLiquidToAlphaList (r_modelsurf_t *ms)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_LiquidBegin;
	al->takedownfunc = R_LiquidEnd;

	// to do - translate midpoint by matrix
	al->type = ALPHATYPE_LIQUID;
	al->dist = R_GetDist (R_TransformMidPoint (ms->surface->midpoint, ms->matrix));
	al->surf = ms->surface;
	al->matrix = ms->matrix;
}


void R_ExtraSurfsBegin (void);
void R_ExtraSurfsEnd (void);

void R_AddSurfaceToAlphaList (r_modelsurf_t *ms)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_ExtraSurfsBegin;
	al->takedownfunc = R_ExtraSurfsEnd;

	// to do - translate midpoint by matrix
	al->type = ALPHATYPE_SURFACE;
	al->dist = R_GetDist (R_TransformMidPoint (ms->surface->midpoint, ms->matrix));
	al->surf = ms->surface;
	al->matrix = ms->matrix;
}


void R_AddFenceToAlphaList (r_modelsurf_t *ms)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	al->setupfunc = R_ExtraSurfsBegin;
	al->takedownfunc = R_ExtraSurfsEnd;

	// to do - translate midpoint by matrix
	al->type = ALPHATYPE_FENCE;
	al->dist = R_GetDist (R_TransformMidPoint (ms->surface->midpoint, ms->matrix));
	al->surf = ms->surface;
	al->matrix = ms->matrix;
}


void R_AddEntityToAlphaList (entity_t *ent)
{
	alphalist_t *al;

	if (!(al = R_GrabAlphaSlot ())) return;

	switch (ent->model->type)
	{
	case mod_alias:
		al->type = ALPHATYPE_ALIAS;
		break;

	case mod_brush:
		al->type = ALPHATYPE_BRUSH;
		break;

	case mod_sprite:
		al->type = ALPHATYPE_SPRITE;
		al->setupfunc = R_SpriteBegin;
		al->takedownfunc = R_SpriteEnd;
		break;

	default:
		Host_Error ("R_AddEntityToAlphaList: unknown model type\n");
		return;
	}

	al->dist = R_GetDist (ent->origin);
	al->ent = ent;
}


int R_AlphaSortFunc (const void *a1, const void *a2)
{
	alphalist_t *al1 = *(alphalist_t **) a1;
	alphalist_t *al2 = *(alphalist_t **) a2;

	// back to front ordering
	return (int) (al2->dist - al1->dist);
}


void R_InvalidateSprite (void);
void R_RenderFenceTexture (msurface_t *surf, float *matrix);
void R_DrawAliasBatches (entity_t **ents, int numents, void *meshbuffer);
void R_SetupAliasModel (entity_t *e);
void R_DrawAlphaBModel (entity_t *ent);
void R_InvalidateLiquid (void);
void R_InvalidateExtraSurf (void);
void R_RenderAlphaSurface (msurface_t *surf, float *m);
void R_RenderLiquidSurface (msurface_t *surf, float *m);


void R_DrawAlphaList (void)
{
	int i;
	alphalist_t *prev;
	alphalist_t *curr;

	if (!r_numalphalist) return;

	if (r_numalphalist > 1)
	{
		qsort
		(
			r_alphalist,
			r_numalphalist,
			sizeof (alphalist_t *),
			R_AlphaSortFunc
		);
	}

	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask (GL_FALSE);

	glEnable (GL_ALPHA_TEST);
	glAlphaFunc (GL_GREATER, 0.005);

	prev = NULL;

	// invalidate all cached states
	R_InvalidateSprite ();
	R_InvalidateLiquid ();
	R_InvalidateExtraSurf ();

	for (i = 0; i < r_numalphalist; i++)
	{
		qboolean newtype = false;

		curr = r_alphalist[i];

		if (!prev)
			newtype = true;
		else if (prev->type != curr->type)
		{
			// invalidate all cached states
			R_InvalidateSprite ();
			R_InvalidateLiquid ();
			R_InvalidateExtraSurf ();

			// callbacks for state takedown (if needed)
			if (prev->takedownfunc)
				prev->takedownfunc ();

			// bring up new (also needed if !prev)
			newtype = true;
		}
		else newtype = false;

		if (newtype && curr->setupfunc)
			curr->setupfunc ();

		// drawing function
		switch (curr->type)
		{
		case ALPHATYPE_ALIAS:
			R_SetupAliasModel (curr->ent);

			if (curr->ent->visframe == r_framecount)
				R_DrawAliasBatches (&curr->ent, 1, scratchbuf);

			break;

		case ALPHATYPE_BRUSH:
			R_DrawAlphaBModel (curr->ent);
			break;

		case ALPHATYPE_SPRITE:
			R_DrawSpriteModel (curr->ent);
			break;

		case ALPHATYPE_PARTICLE:
			R_DrawParticlesForType (curr->particles);
			break;

		case ALPHATYPE_FENCE:
			R_RenderFenceTexture (curr->surf, curr->matrix);
			break;

		case ALPHATYPE_LIQUID:
			R_RenderLiquidSurface (curr->surf, curr->matrix);
			break;

		case ALPHATYPE_SURFACE:
			R_RenderAlphaSurface (curr->surf, curr->matrix);
			break;

		case ALPHATYPE_CORONA:
			R_RenderCorona (curr->light);
			break;

		default:
			break;
		}

		prev = curr;
	}

	if (prev->takedownfunc)
		prev->takedownfunc ();

	glColor4f (1, 1, 1, 1);	// mh - current color is undefined after vertex arrays so this is always needed
	glAlphaFunc (GL_GREATER, 0.666);
	glDisable (GL_ALPHA_TEST);

	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	// not strictly speaking necessary but it does the job well enough
	r_numalphalist = 0;
}


/*
====================
GL_Fullbrights_f -- johnfitz
====================
*/
void GL_Fullbrights_f (void)
{
	TM_ReloadNobrightImages ();
}

/*
====================
R_SetClearColor_f -- johnfitz
====================
*/
void R_SetClearColor_f (void)
{
	byte	*rgb;
	int		s;

	s = (int) r_clearcolor.value & 0xFF;
	rgb = (byte *) (d_8to24table + s);
	glClearColor (rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0, 0);
}

/*
====================
R_Novis_f -- johnfitz
====================
*/
void R_Novis_f (void)
{
	extern int vis_changed;
	vis_changed = 1;
}

/*
====================
R_OldSkyLeaf_f -- johnfitz
====================
*/
void R_OldSkyLeaf_f (void)
{
	extern int vis_changed;
	vis_changed = 1;
}


/*
===============
R_Envmap_f

Grab six views for environment mapping tests
===============
*/
void R_Envmap_f (void)
{
}

/*
===============
R_Init
===============
*/
void R_InitQuads (void);
void R_WarpRegisterCvars (void);

void R_Init (void)
{
	extern cvar_t r_show_coronas;
	extern cvar_t gl_finish;
	extern cvar_t gl_subdivide_size;
	extern cvar_t r_motionblur;
	static byte data[16][4];
	int i;

	R_InitQuads ();

	Cmd_AddCommand ("timerefresh", R_TimeRefresh_f);
	Cmd_AddCommand ("envmap", R_Envmap_f);
	Cmd_AddCommand ("pointfile", R_ReadPointFile_f);

	R_WarpRegisterCvars ();

	Cvar_RegisterVariable (&r_motionblur, NULL);
	Cvar_RegisterVariable (&r_show_coronas, NULL);
	Cvar_RegisterVariable (&r_showtris, NULL);
	Cvar_RegisterVariable (&r_norefresh, NULL);
	Cvar_RegisterVariable (&r_lightmap, NULL);
	Cvar_RegisterVariable (&r_fullbright, NULL);
	Cvar_RegisterVariable (&r_drawentities, NULL);
	Cvar_RegisterVariable (&r_drawviewmodel, NULL);
	Cvar_RegisterVariable (&r_shadows, NULL);
	Cvar_RegisterVariable (&r_wateralpha, NULL);
	Cvar_RegisterVariable (&r_dynamic, NULL);
	Cvar_RegisterVariable (&r_novis, R_Novis_f);
	Cvar_RegisterVariable (&r_speeds, NULL);

	Cvar_RegisterVariable (&gl_finish, NULL);
	Cvar_RegisterVariable (&gl_clear, NULL);
	Cvar_RegisterVariable (&gl_cull, NULL);
	Cvar_RegisterVariable (&gl_smoothmodels, NULL);
	Cvar_RegisterVariable (&gl_affinemodels, NULL);
	Cvar_RegisterVariable (&gl_polyblend, NULL);
	Cvar_RegisterVariable (&gl_playermip, NULL);
	Cvar_RegisterVariable (&gl_nocolors, NULL);

	// johnfitz -- new cvars
	Cvar_RegisterVariable (&r_clearcolor, R_SetClearColor_f);
	Cvar_RegisterVariable (&r_waterwarp, NULL);
	Cvar_RegisterVariable (&r_drawflat, NULL);
	Cvar_RegisterVariable (&r_flatlightstyles, NULL);
	Cvar_RegisterVariable (&r_oldskyleaf, R_OldSkyLeaf_f);
	Cvar_RegisterVariable (&r_drawworld, NULL);
	Cvar_RegisterVariable (&r_showbboxes, NULL);
	Cvar_RegisterVariable (&gl_fullbrights, GL_Fullbrights_f);
	Cvar_RegisterVariable (&gl_overbright, NULL);
	Cvar_RegisterVariable (&gl_overbright_models, NULL);
	Cvar_RegisterVariable (&r_lerpmodels, NULL);
	Cvar_RegisterVariable (&r_lerpmove, NULL);
	Cvar_RegisterVariable (&r_nolerp_list, R_NoLerpList_f);
	// johnfitz

	Cvar_RegisterVariable (&gl_src_blend, NULL);
	Cvar_RegisterVariable (&gl_dst_blend, NULL);

	Cvar_RegisterVariable (&r_lockpvs, NULL);
	Cvar_RegisterVariable (&r_lockfrustum, NULL);

	Cvar_RegisterVariable (&gl_subdivide_size, NULL); // johnfitz -- moved here from gl_model.c

	R_InitParticles ();
	R_SetClearColor_f (); // johnfitz

	Sky_Init (); // johnfitz
	Fog_Init (); // johnfitz

	for (i = 0; i < 16; i++)
	{
		data[i][0] = 128;
		data[i][1] = 128;
		data[i][2] = 128;
		data[i][3] = 255;
	}

	r_whitetexture = TM_LoadImage (NULL, "", 4, 4, SRC_RGBA, (byte *) data, "", (src_offset_t) data, TEXPREF_NONE);

#ifdef GLTEST
	Test_Init ();
#endif
}

/*
===============
R_NoLerpList_f -- johnfitz -- called when r_nolerp_list cvar changes
===============
*/
void R_NoLerpList_f (void)
{
	int i;

	for (i = 0; i < MAX_MODELS; i++)
		Mod_SetExtraFlags (cl.model_precache[i]);
}

/*
===============
R_TranslatePlayerSkin -- johnfitz -- rewritten.  also, only handles new colors, not new skins
===============
*/
void R_TranslatePlayerSkin (int playernum)
{
	int			top, bottom;

	top = (cl.scores[playernum].colors & 0xf0) >> 4;
	bottom = cl.scores[playernum].colors & 15;

	//FIXME: if gl_nocolors is on, then turned off, the textures may be out of sync with the scoreboard colors.
	if (!gl_nocolors.value)
		if (playertextures[playernum])
			TM_ReloadImage (playertextures[playernum], top, bottom);
}

/*
===============
R_TranslateNewPlayerSkin -- johnfitz -- split off of TranslatePlayerSkin -- this is called when
the skin or model actually changes, instead of just new colors
added bug fix from bengt jardup
===============
*/
void R_TranslateNewPlayerSkin (int playernum)
{
	char		name[64];
	byte		*pixels;
	aliashdr_t	*paliashdr;
	int			skinnum;
	entity_t	*e;

	//get correct texture pixels
	e = cl_entities[1+playernum];

	if (!e->model || e->model->type != mod_alias)
		return;

	paliashdr = (aliashdr_t *) Mod_Extradata (e->model);

	skinnum = e->skinnum;

	//TODO: move these tests to the place where skinnum gets received from the server
	if (skinnum < 0 || skinnum >= paliashdr->numskins)
	{
		Con_DPrintf ("(%d): Invalid player skin #%d\n", playernum, skinnum);
		skinnum = 0;
	}

	pixels = (byte *) paliashdr + paliashdr->texels[skinnum]; // This is not a persistent place!

	//upload new image
	sprintf (name, "player_%i", playernum);
	playertextures[playernum] = TM_LoadImage (e->model, name, paliashdr->skinwidth, paliashdr->skinheight,
								SRC_INDEXED, pixels, paliashdr->gltextures[skinnum][0]->source_file, paliashdr->gltextures[skinnum][0]->source_offset, TEXPREF_PAD | TEXPREF_OVERWRITE);

	//now recolor it
	R_TranslatePlayerSkin (playernum);
}

/*
===============
R_NewGame -- johnfitz -- handle a game switch
===============
*/
void R_NewGame (void)
{
	int i;

	//clear playertexture pointers (the textures themselves were freed by texmgr_newgame)
	for (i = 0; i < MAX_SCOREBOARD; i++)
		playertextures[i] = NULL;
}

/*
===============
R_NewMap
===============
*/
// mh 2010-07-16 - clear down modelsurfs from the previous map
void R_ModelSurfsBeginMap (void);
void WaterAlpha_ParseWorldspawn (void);
void R_CreateMeshIndexBuffer (void);
void R_InitVertexBuffers (void);

void R_NewMap (void)
{
	int		i;
	int start, end;

	start = SDL_GetTicks ();

	for (i = 0; i < 256; i++)
		d_lightstylevalue[i] = 264;		// normal light value

	// mh 2010-07-16 - clear down stuff from the previous map
	R_ModelSurfsBeginMap ();
	R_AlphaListBeginMap ();

	// create any static vertex and index buffers we need for this map
	R_CreateMeshIndexBuffer ();

	// clear out efrags in case the level hasn't been reloaded
	// FIXME: is this one short?
	for (i = 0; i < cl.worldmodel->numleafs; i++)
		cl.worldmodel->leafs[i].efrags = NULL;

	r_viewleaf = NULL;
	R_ClearParticles ();

	GL_BuildLightmaps ();

	r_framecount = 0; // johnfitz -- paranoid?
	r_visframecount = 0; // johnfitz -- paranoid?

	Sky_NewMap (); // johnfitz -- skybox in worldspawn
	Fog_NewMap (); // johnfitz -- global fog in worldspawn
	// WaterAlpha_ParseWorldspawn ();

	R_InitVertexBuffers ();

	end = SDL_GetTicks ();

	Con_DPrintf ("R_NewMap in %i ms\n", (end - start));
}


/*
====================
R_TimeRefresh_f

For program optimization
====================
*/
void R_TimeRefresh_f (void)
{
}

void D_FlushCaches (void)
{
}
